from collections import defaultdict
from typing import NamedTuple, Tuple, List, Mapping

import funcy as fn
from bidict import bidict

import aiger
import aiger.common as cmn


class CNF(NamedTuple):
    clauses: List[Tuple[int]]
    symbol_table: Mapping[str, int]
    max_var: int

    def id2name(self, idx):
        return self.symbol_table.inv[abs(idx)]

    def name2id(self, name):
        return self.symbol_table[name]


def aig2cnf(circ, output=None):
    """Convert an AIGER circuit to CNF via the Tseitin transformation."""
    circ = circ.aig  # Extract AIG from potential wrapper.
    assert len(circ.latches) == 0
    if output is None:
        assert len(circ.outputs) == 1
        output = fn.first(circ.outputs)

    max_var = 0

    def fresh():
        nonlocal max_var
        max_var += 1
        return max_var

    output = dict(circ.node_map)[output]
    symbol_table = defaultdict(fresh)  # maps input names to tseitin variables

    clauses, gates = [], {}  # maps gates to tseitin variables
    for gate in cmn.eval_order(circ):
        if isinstance(gate, aiger.aig.ConstFalse):
            true_var = symbol_table[gate]
            clauses.append((true_var,))
            gates[gate] = -true_var
        elif isinstance(gate, aiger.aig.Inverter):
            gates[gate] = -gates[gate.input]
        elif isinstance(gate, aiger.aig.Input):
            gates[gate] = symbol_table[gate.name]
        elif isinstance(gate, aiger.aig.AndGate):
            gates[gate] = fresh()
            clauses.append((-gates[gate.left], -gates[gate.right],  gates[gate]))  # noqa
            clauses.append((gates[gate.left],                     -gates[gate]))  # noqa
            clauses.append((                    gates[gate.right], -gates[gate]))  # noqa

    clauses.append((gates[output],))
    return CNF(clauses, bidict(symbol_table), max_var)



from itertools import product

import aiger
import hypothesis.strategies as st
from aiger import hypothesis as aigh
from hypothesis import given
from pysat.solvers import Glucose3

from aiger_cnf import aig2cnf


def gen_tests(new_vars, expr2, val):
    for inputs in product(*((True, False),)*len(new_vars)):
        inputs = dict(zip(new_vars, inputs))
        inputs.update(val)
        yield expr2(inputs)


@given(aigh.Circuits, st.data())
def test_aig2cnf(circ, data):
    expr1 = aiger.BoolExpr(circ.unroll(1))
    cnf = aig2cnf(expr1)
    g = Glucose3()
    for c in cnf.clauses:
        g.add_clause(c)

    test_input = {i: data.draw(st.booleans()) for i in expr1.inputs}
    assumptions = []
    for name, val in test_input.items():
        if name not in cnf.symbol_table:
            continue
        sym = cnf.symbol_table[name]
        if not val:
            sym *= -1
        assumptions.append(sym)

    assert expr1(test_input) == g.solve(assumptions=assumptions)
